
import argparse
from pathlib import Path
from typing import List

import pydot
import networkx as nx

from . import graph_viz_builder
from .utils import paths2graph


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=str, help="Path to directory containing python project.")
    #
    parser.add_argument("--dir-as", type=str, default="node", help="Draw a directory as 'node' (default) or 'cluster'.")
    parser.add_argument("--output-file", type=str, help="Replace dot string output with name of image file incl. extension (e.g. img.png). Supports file formats from GraphViz (e.g. png, svg).")
    parser.add_argument("--ignore-files", type=str, help="Glob pattern for ignoring files.")
    # layout components
    parser.add_argument("--show-interface", action="store_true", help="Show interface (var + function names) on file nodes.")
    parser.add_argument("--show-imports", action="store_true", help="Show import links between files.")
    parser.add_argument("--show-class-bases", action="store_true", help="Show the base classes for a class.")
    parser.add_argument("--show-func-return-type", action="store_true", help="Show the return type of a function.")
    parser.add_argument("--show-func-decorators", action="store_true", help="Show the decorators of a function.")
    # styling
    parser.add_argument("--import-link-color", type=str, default="black", help="Color of import link, e.g. 'black', 'gray', 'royalblue'. See options: https://www.graphviz.org/doc/info/colors.html")
    args = parser.parse_args()

    # Make args available to the graph_viz_builder module.
    # Purpose: easy access to styling options, e.g. import-link-color.
    # Otherwise all these options would have to be pipped all the way down.
    graph_viz_builder.ARGS = args

    # must be a directory path, for now
    path = Path(args.path)
    if not path.is_dir():
        raise argparse.ArgumentTypeError("Must be a directory path")
    
    # find all python files in path
    filepaths: List[Path] = list(path.rglob("*.py"))

    # optional: ignore files pattern
    ignore_filepaths = []
    if args.ignore_files:
        ignore_filepaths: List[Path] = list(path.rglob(args.ignore_files))
    filepaths = [p for p in filepaths if p not in ignore_filepaths]

    # convert filepaths to graph
    graph: nx.DiGraph = paths2graph.to_graph(paths=filepaths)

    # create directory view
    dot: pydot.Dot = graph_viz_builder.build_dot_layout(g=graph, 
                                                    project_path=path,
                                                    dir_as=args.dir_as, 
                                                    show_interface=args.show_interface,
                                                    show_imports=args.show_imports
                                                    )
    
    # output as string or PNG file
    if args.output_file:
        file_ext = args.output_file.split(".")[-1]
        dot.write(args.output_file, prog="dot", format=file_ext)
    else:
        print(dot.to_string())




from typing import List
from pathlib import Path
import argparse
import re

import networkx as nx
import pydot
import astroid

from .module.interface import ModuleInterface
from .module import interface
from .module.imports import get_module_imports


# Global reference to program args, assigned from outside this module.
ARGS: argparse.Namespace = None


class EdgeLayout(pydot.Edge):
    """ Dot layout for an edge.
    """
    def __init__(self, src: Path, dst: Path) -> None:
        super().__init__(
            src=pydot.Node(src.as_posix()), 
            dst=pydot.Node(dst.as_posix()),
            color="gray",
            style="solid"
            )


class DirLayout(pydot.Node):
    """ Dot layout for a directory node.
    """
    def __init__(self, graph: nx.DiGraph, node: Path) -> None:
        super().__init__(name=node.as_posix())
        if _is_package(graph=graph, node=node): self.set("shape", "component")
        else: self.set("shape", "folder")
        self.set("color", "red")
        self.set("label", node.relative_to(node.parent).as_posix())


class FileLayout(pydot.Node):
    """ Dot layout for a file node.
    """
    def __init__(self, node: Path, with_interface: bool) -> None:
        super().__init__(name=node.as_posix())
        self.set("shape", "record")
        filename = node.relative_to(node.parent).as_posix()
        if with_interface:
            mod_int: ModuleInterface = interface.get_module_interface(module_path=node)
            cs: List[str] = self._class_definition_style(mod_int.class_defs)
            vs: List[str] = [v.name for v in mod_int.single_assignments]
            fs: List[str] = self._function_def_style(mod_int.function_defs)
            
            layout = f"{{ {filename}| {FileLayout.to_record_str(cs + vs)} | {FileLayout.to_record_str(fs)} }}"
            self.set(
                "label", 
                layout
                )
        else:
            self.set("label", filename)

    def _class_definition_style(self, class_defs: List[astroid.ClassDef]) -> List[str]:
        """ Format class names with or without it's bases.
                E.g. 'Class1()' or 'Class1(Foo, Bar)'
        """
        classes_with_bases = []
        bases = []
        for c in class_defs:
            if ARGS.show_class_bases:
                bases = [base.as_string() for base in c.bases]
            classes_with_bases.append( f"{c.name}({', '.join(bases)})")
        return classes_with_bases

    def _function_def_style(self, function_defs: List[astroid.FunctionDef]) -> List[str]:
        """ Format function names with or without return type string.
                E.g. 'func' or 'func: int'
        """
        # return type string pattern
        # match group between '->' (remove whitespace) and ':' 
        # e.g. def bla(a,b) -> int: ... --> 'int'
        pattern = "->\s*(.*):"
        fs = []
        for func_def in function_defs:
            # NOTE: decorators are added as a new "function" layout-wise before the function,
            #       ergo this has to be first.
            # if show decorators AND has decorators
            if ARGS.show_func_decorators and func_def.decorators:
                deco_strings = []
                for d in func_def.decorators.nodes:
                    deco_strings.append(d.as_string())
                deco_string = "@ " + ", ".join(deco_strings)
                fs.append(deco_string)
            
            func_name = func_def.name
            # if show return type AND function has a defined return type
            if ARGS.show_func_return_type and func_def.returns:
                return_type_str = re.findall(pattern, func_def.as_string())[0]
                func_name += f": {return_type_str}"

            fs.append(func_name)
        return fs

    @staticmethod
    def to_record_str(lst: List[str]) -> str:
        """ Convert a list of strings to a single string in the dot record format.
            E.g.: attr1\lattr2\l --> attr1 | attr2
        """
        return "\l".join(lst) + "\l"



def _is_package(graph: nx.DiGraph, node: Path) -> bool:
    """ if a node has a successor that is an __init__.py file
    """
    s: Path
    for s in graph.successors(node):
        if s.relative_to(node).as_posix() == "__init__.py":
            return True
    return False


def _is_file(graph: nx.DiGraph, node: Path) -> bool:
    """ A leaf node, i.e. with no successor nodes, is a file.
    """
    return len(list(graph.successors(node))) == 0


def _parse_python_file(filepath: Path) -> astroid.Module:
    """ Parse a python file with astroid.
    """
    with open(filepath) as fh:
            code: str = fh.read()
    return astroid.parse(code)


def _handle_weird_pydot_name(cluster_node: Path) -> str:
    """ pydot.Node names appear to be wrapped in an extra layer of strings
            e.g. "dir1/file1.py" --> '"dir1/file1.py"'
        So, pass string through Node constructor and retrieve name, in order to replicate.
    """
    cluster_node_name = pydot.Node(name=cluster_node.as_posix()).get_name()
    return cluster_node_name


class GraphVizBuilder:
    def __init__(self, 
                 path_graph: nx.DiGraph, 
                 project_path: Path, 
                 rankdir: str = "TB"
                 ) -> None:
        self.rankdir = rankdir
        self.g = path_graph
        self.project_path = project_path

        self.file_nodes = [n for n in self.g.nodes if _is_file(graph=self.g, node=n)]
        self.dir_nodes = [n for n in self.g.nodes if not _is_file(graph=self.g, node=n)]

        self.reset()

    def reset(self) -> None:
        self._graph = pydot.Dot(graph_type="digraph", 
                                rankdir=self.rankdir,
                                fontname="Helvetica"
                                )
        self._graph.set_node_defaults(fontname="Helvetica")

    @property
    def graph(self) -> pydot.Dot:
        product = self._graph
        self.reset()
        return product

    def add_file_nodes(self, with_interface: bool = False) -> None:
        for n in self.file_nodes:
            node = FileLayout(node=n, with_interface=with_interface)
            self._graph.add_node(node)

    def add_dir_nodes(self) -> None:
        for n in self.dir_nodes:
            node = DirLayout(graph=self.g, node=n)
            self._graph.add_node(node)

    def add_dir_clusters(self) -> None:
        for n in self.dir_nodes:
            c = pydot.Cluster(n.as_posix(), 
                            #  label=n.relative_to(n.parent).as_posix(), 
                             label=n.as_posix(),
                             color="gray")
            # add nodes to cluster
            cluster_nodes = [s for s in self.g.successors(n) if _is_file(graph=self.g, node=s)]

            # find cluster nodes in _graph instead of g
            for c_node in cluster_nodes:
                c_node_name = _handle_weird_pydot_name(cluster_node=c_node)
                # returns a list, get first item 
                node_list = self._graph.get_node( c_node_name )
                assert len(node_list) == 1, "only one node should exist with given name"
                node = node_list[0]
                c.add_node(node)

            # add cluster to graph
            self._graph.add_subgraph(c)


    def add_hierarchy_links(self) -> None:
        for n in self.g.nodes:
            for s in self.g.successors(n):
                edge = EdgeLayout(src=n, dst=s)
                self._graph.add_edge(edge)

    def add_import_links(self) -> None:
        for n in self.file_nodes:
            module_ast = _parse_python_file(n)
            internal_imports, _ = get_module_imports(module_path=n, module_ast=module_ast, project_path=self.project_path)
            for int_imp in internal_imports:
                edge = pydot.Edge(
                    src=pydot.Node(n.as_posix()), 
                    dst=pydot.Node(int_imp.as_posix()),
                    color=ARGS.import_link_color,
                    style="dashed"
                )
                self._graph.add_edge(edge)

    def add_hightlight(self) -> None:
        n: pydot.Node
        for n in self._graph.get_node_list():
            print(n.get_name())
            n.set_style("filled")
            n.set_fillcolor("green")
        

def build_dot_layout(g: nx.DiGraph, 
                     project_path: Path, 
                     dir_as: str = "node",
                     show_interface: bool = False,
                     show_imports: bool = False
                     ) -> pydot.Dot:
    builder = GraphVizBuilder(path_graph=g, project_path=project_path)
    builder.add_file_nodes(with_interface=show_interface)

    if dir_as == "node":
        builder.add_dir_nodes()
        builder.add_hierarchy_links()
    elif dir_as == "cluster":
        builder.add_dir_clusters()
    else:
        raise ValueError(f"dir_as cannot take value: {dir_as}")
    
    if show_imports:
        builder.add_import_links()

    # builder.add_hightlight()

    return builder.graph




from typing import List, Dict, Tuple
from dataclasses import dataclass
from pathlib import Path

import astroid


@dataclass
class ModuleInterface:
    class_defs:         List[astroid.ClassDef]
    function_defs:      List[astroid.FunctionDef]
    single_assignments: List[astroid.AssignName]


def get_single_assignments(m: astroid.Module) -> List[astroid.AssignName]:
    """ Returns single assignments in top scope of a module.
        Only include single assignment.
        E.g. "a = 11", where the first element/child if of type AssignName (= "a").
        Ignore tuple case: "a, b = some_func()".
    """ 
    assigns = [e for e in m.body if isinstance(e, astroid.Assign)]
    single_assigns = [list(a.get_children())[0] for a in assigns if isinstance(list(a.get_children())[0], astroid.AssignName)]
    return single_assigns


def main(module: astroid.Module) -> ModuleInterface:
    return ModuleInterface(
        class_defs=[e for e in module.body if isinstance(e, astroid.ClassDef)],
        function_defs=[e for e in module.body if isinstance(e, astroid.FunctionDef)],
        single_assignments=get_single_assignments(m=module)
    )


def get_module_interface(module_path: Path) -> ModuleInterface:
    with open(module_path) as fh:
        code: str = fh.read()
    m = astroid.parse(code)
    return main(module=m)




import pathlib
from pathlib import Path, PurePath
from typing import Any, List, Dict, Optional, Tuple, Union
from functools import singledispatch

import astroid
import networkx as nx


class BaseImportPath:
    def __init__(self, path) -> None:
        assert not path.suffix, "An import path should not point to a specific file."
        self.path = pathlib.Path(path)
    
    def __getattr__(self, name: str) -> Any:
        return getattr(self.path, name)

    def __str__(self) -> str:
        return str(self.path)

    def __truediv__(self, key):
        return self.path.__truediv__(key)

    def try_to_qualify(self):
        raise NotImplementedError

    @staticmethod
    def try_to_qualify_import_path(import_path: Path) -> Optional[Path]:
        # imports a module, if it points to a .py file with same name
        module_import = Path( str(import_path) + ".py" )
        # imports a package, if it points to a dir with a __init__.py file
        package_import = import_path / "__init__.py"
        if module_import.exists(): return module_import
        if package_import.exists(): return package_import
        # could not be qualified / linked to a specific .py file
        return None


class ImportPath(BaseImportPath):
    def __init__(self, path) -> None:
        super().__init__(path)

    def try_to_qualify(self) -> Optional[Path]:
        """ Import can import: module or package, but not element in either.
            Examples:
                package/module --> package/module.py
                package --> package/__init__.py
        """
        qual_import = BaseImportPath.try_to_qualify_import_path(self.path)
        if qual_import and qual_import.suffix == ".py":
            return qual_import
        return None


class ImportFromPath(BaseImportPath):
    def __init__(self, path) -> None:
        super().__init__(path)
    
    def try_to_qualify(self) -> Optional[Path]:
        """ ImportFrom can import: package, module or element in package or module
            Examples:
                package/module/func --> package/module.py
                package/module --> package/module.py
                package --> package/__init__.py
        """
        # if imports a module or package
        qual_import = BaseImportPath.try_to_qualify_import_path(self.path)
        if qual_import and qual_import.suffix == ".py":
            return qual_import

        # if imports an element in a module or package
        # try to qualify import path parent
        qual_import = BaseImportPath.try_to_qualify_import_path(self.path.parent)
        if qual_import and qual_import.suffix == ".py":
            return qual_import
        # import from path could not be qualified
        return None


def _to_rel_importfrom_paths(import_from: astroid.ImportFrom, module_path: Path) -> List[ImportFromPath]:
    """ Convert ImportFrom statement to import paths relative to module path.
        Ex: 
            import_from: from ..modname import name1 as n1, name2
            module_path: dir1/dir2/module.py
            --> dir1/modname
            --> dir1/modname/name1 (alias ignored)
                dir1/modname/name2
    """
    assert import_from.level, "level must be present for a relative import"
    p = module_path.parents[import_from.level - 1] / import_from.modname.replace(".", "/")
    return [ImportFromPath(p / n[0]) for n in import_from.names]


def _to_abs_importfrom_paths(import_from: astroid.ImportFrom, project_path: Path) -> List[ImportFromPath]:
    assert not import_from.level, "level must NOT be present for an absolute import"
    path_strs = [(import_from.modname + "." + name).replace(".", "/") for name,alias in import_from.names]
    return [ImportFromPath(project_path / Path(p)) for p in path_strs]


def _to_abs_import_paths(abs_import: astroid.Import, project_path: Path) -> List[ImportPath]:
    return [ImportPath(project_path / name.replace(".", "/")) for name,alias in abs_import.names]

    
def relative_import_from(module: astroid.Module, module_path: Path) -> List[ImportFromPath]:
    """ Returns the relative ImportFrom's of the module.
        Ex: relative ImportFrom
            from .mod1 import func1, func2
            from . import mod1, mod2
    """
    rel_import_froms = [e for e in module.body if isinstance(e, astroid.ImportFrom) and e.level]

    import_paths: List[ImportPath] = []
    for import_from in rel_import_froms:
        import_paths.extend( _to_rel_importfrom_paths(import_from, module_path) )
    return import_paths


def absolute_import_from(module: astroid.Module, project_path: Path) -> List[ImportPath]:
    """ Returns the absolute ImportFrom's of the module.
        Ex: absolute ImportFrom
            from package1.mod import class, func
            from package1 import mod1 as m1, mod2
    """
    abs_import_froms = [e for e in module.body if isinstance(e, astroid.ImportFrom) and not e.level]

    import_paths = []
    for imp in abs_import_froms:
        import_paths.extend( _to_abs_importfrom_paths(import_from=imp, project_path=project_path) )
    return import_paths


def absolute_import(module: astroid.Module, project_path: Path) -> List[ImportPath]:
    """ Returns the absolute Import of the module.
        Ex: absolute Import
            import mod
    """
    imports = [e for e in module.body if isinstance(e, astroid.Import)]

    import_paths = []
    for imp in imports:
        import_paths.extend( _to_abs_import_paths(abs_import=imp, project_path=project_path) )
    return import_paths


def qualify_import_paths(import_paths: List[BaseImportPath]) -> Tuple[List[Path], List[BaseImportPath]]:
    qual_paths = []
    non_qual_paths = []
    for p in import_paths:
        maybe_qual = p.try_to_qualify()
        if maybe_qual: qual_paths.append(maybe_qual)
        else: non_qual_paths.append(p)
    return qual_paths, non_qual_paths


def is_dir(graph: nx.DiGraph, node: Path) -> bool:
    return len(list(graph.successors(node))) > 0


def get_module_imports(module_path: Path, 
                       module_ast: astroid.Module,
                       project_path: Path
                       ) -> Tuple[List[Path], List[Path]]:
    # --- ImportFrom
    # Import paths that do NOT point to specific python files
    rel_import_from: List[Path] = relative_import_from(module=module_ast, module_path=module_path)
    abs_import_from: List[Path] = absolute_import_from(module=module_ast, project_path=project_path)
    # --- Import
    abs_import: List[Path] = absolute_import(module=module_ast, project_path=project_path)

    all_import_paths = abs_import + abs_import_from + rel_import_from
    qual_import_paths, non_qual_import_paths = qualify_import_paths(import_paths=all_import_paths)

    for p in qual_import_paths:
        assert p.suffix == ".py", "all qualified paths must point to a .py file"

    for p in non_qual_import_paths:
        assert not p.suffix, "all non-qualifed paths must not point to a specific file"

    # only show top-level package name for external imports
    # e.g. "from sklearn.mixtures import GMM" --> sklearn
    ext_toplevel_imports = [Path(ext.relative_to(project_path).parts[0]) for ext in non_qual_import_paths]

    # remove duplicates
    internal_imports = qual_import_paths
    return list(set(internal_imports)), list(set(ext_toplevel_imports))



""" Converts a set of Paths to a NetworkX graph
"""

from typing import Iterable, List, Tuple
from pathlib import Path, PureWindowsPath

import networkx as nx

# import streamlit as st

def _edges_from_path(path: Path) -> List[Tuple]:
    # path and parents
    # pp = ([path] + list(path.parents)[:-1])[::-1]
    pp = ([path] + list(path.parents))[::-1]
    edges = [(pp[i], pp[i+1]) for i in range(len(pp)-1)]
    return edges


def _add_edges_as_weighted(g, edges) -> None:
    for (src,dst) in edges:
        if g.has_edge(src, dst): 
            g[src][dst]["weight"] += 1
        else: 
            g.add_edge(src, dst, weight=1)


def to_graph(paths: Iterable[Path], is_windows_paths=False) -> nx.DiGraph:
    """
    """
    if is_windows_paths:
        paths = [PureWindowsPath(p) for p in paths]
    g = nx.DiGraph()

    all_edges = []
    for path in paths:
        edges = _edges_from_path(path)
        all_edges.extend(edges)
    _add_edges_as_weighted(g, all_edges)
    return g


from typing import List
from setuptools import setup, find_packages


setup(
    name='moduml',
    description='UML-inspired diagrams of python modules and packages.',
    url='https://github.com/roschly/moduml',
    download_url='https://github.com/roschly/moduml/archive/refs/tags/v0.1.1.tar.gz',
    author='roschly',
    entry_points={
        'console_scripts': [
            'moduml = moduml.__main__:main',
        ]
    },
    packages=find_packages(),
    install_requires=[
        'astroid',
        'networkx',
        'pydot',
    ]
)